//
//  ViewController.m
//  pubkey-pin-ios
//
//  Created by Jeffrey Walton on 2/1/13.
//  Copyright (c) 2013 OWASP. All rights reserved.

//
// A lot of this was shameless ripped from
// http://blog.securemacprogramming.com/2011/12/on-ssl-pinning-for-cocoa-touch/.
// Unfortunately, we are pinning a certifcate, and not a public key. (Windows and Android
// samples do pin the public key). This will be a problem for sites like Google, which
// rotate the 'outer' certificate every 30 days or so. While Google rotates the X509
// certifcate attributes, the 'inner' key (SubjectPublicKeyInfo) remains static.

#import "ViewController.h"

#include "PinAssert.h"

@interface ViewController ()

@end

@implementation ViewController

@synthesize m_fetchedData;
@synthesize m_fetchedLabel, m_fetchButton;

#pragma mark Error reporting

-(void)displayError:(NSString*)msg
{
    UIAlertView* alert = [[UIAlertView alloc] initWithTitle:@"Public Key Pinning"
                                                    message:msg
                                                   delegate:self
                                          cancelButtonTitle:@"Ok"
                                          otherButtonTitles: nil];
    
    ASSERT(nil != alert);
    if(nil != alert)
    {
        [alert show];
        [alert release];
    }
}

#pragma mark UI interactions

-(IBAction)fetchButtonTapped:(id)sender
{
    ASSERT(nil != sender);
    ASSERT(nil != m_fetchedLabel);
    
    [m_fetchedLabel setText:@""];

    NSString* requestString = @"https://www.random.org/integers/?num=16&min=0&max=255&col=16&base=16&format=plain&rnd=new";
    ASSERT(nil != requestString);
    if(!(nil != requestString))
    {
        [self displayError:@"Failed to create requestString."];
        return;
    }
    
    NSURL* requestUrl = [NSURL URLWithString:requestString];
    ASSERT(nil != requestUrl);
    if(!(nil != requestUrl))
    {
        [self displayError:@"Failed to create requestString."];
        return;
    }
    
    NSURLRequest* request = [NSURLRequest requestWithURL:requestUrl
                                             cachePolicy:NSURLRequestReloadIgnoringLocalCacheData
                                         timeoutInterval:10.0f];
    ASSERT(nil != request);
    if(!(nil != request))
    {
        [self displayError:@"Failed to create request."];
        return;
    }
    
    // Note that the delegate for the NSURLConnection is self, so delegate methods must be defined in this file
    NSURLConnection* connection = [[NSURLConnection alloc] initWithRequest:request delegate:self];
    ASSERT(nil != connection);
    if(!(nil != connection))
    {
        [self displayError:@"Failed to create connection."];
        return;
    }
    
    self.m_fetchedData = [[NSMutableString string] retain];
    ASSERT(nil != m_fetchedData);
}

#pragma mark UIView methods

- (void)viewDidLoad
{
    [super viewDidLoad];
	
    // ASSERT(nil != m_fetchedData);
    ASSERT(nil != m_fetchButton);
}

#pragma mark Memory methods

-(void)viewDidUnload
{
    self.m_fetchButton = nil;
    self.m_fetchedLabel = nil;
}

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];
    // Dispose of any resources that can be recreated.
}

- (void) dealloc
{
    if(nil != m_fetchButton)
        [m_fetchButton release];
    
    if(nil != m_fetchedLabel)
        [m_fetchedLabel release];
    
    m_fetchButton = nil;
    m_fetchedLabel = nil;
    
    [super dealloc];
}

#pragma mark NSURLConnection methods

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response
{
    // This method is called when the server has determined that it
    // has enough information to create the NSURLResponse.
	
    // It can be called multiple times, for example in the case of a
    // redirect, so each time we reset the data.
    
    ASSERT(nil != connection);
    ASSERT(nil != response);
    ASSERT(nil != m_fetchedData);
    
    // This could happen if the button is tapped with outstanding fetch requests
    if(nil == m_fetchedData)
        self.m_fetchedData = [[NSMutableString string] retain];
    
    [m_fetchedData setString:@""];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{
    // Append the new data to m_fetchedData.
    
    ASSERT(nil != connection);
    ASSERT(nil != data);
    ASSERT(nil != m_fetchedData);
    
    if(nil == m_fetchedData || nil == data) return;
    
    NSString* t = [[NSString alloc] initWithBytes:[data bytes] length:[data length]
                                         encoding:NSASCIIStringEncoding];
    [m_fetchedData appendString:t];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
    // release the connection, and the data object
    // m_fetchedData is declared as a method instance elsewhere
    
    ASSERT(nil != connection);
    ASSERT(nil != error);
    ASSERT(nil != m_fetchedData);
    
    if(nil != connection)
        [connection release];
    
    if(nil != m_fetchedData)
        self.m_fetchedData  = nil;
    
    [self displayError:[error localizedDescription]];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection
{
    ASSERT(nil != connection);
    ASSERT(nil != m_fetchedData);
    ASSERT(nil != m_fetchedLabel);
    
    do
    {
        NSArray* values = [m_fetchedData componentsSeparatedByCharactersInSet:
                           [NSCharacterSet whitespaceAndNewlineCharacterSet]];

        ASSERT(nil != values);
        if(!(nil != values)) break;
        
        NSMutableString* display = [NSMutableString stringWithCapacity:(([values count])*5 + 2)];
        ASSERT(nil != display);
        if(!(nil != display)) break;
        
        for(NSUInteger i = 0; i < [values count]; i++)
        {
            id v = values[i];
            ASSERT(v != nil);
            
            NSString* s = nil;
            if([v isKindOfClass:[NSString class]])
                s = (NSString*)v;
            
            if(nil != s && [s length] > 0)
                [display appendString:[NSString stringWithFormat:@"0x%@ ", s]];
            
            if(0 != i && (i+1)%4 == 0)
                [display appendString:@"\n"];
        }
        
        ASSERT([display length] > 0);
        if(nil != m_fetchedLabel)
            [m_fetchedLabel setText:display];
        
    } while (0);
    
    if(nil != connection)
        [connection release];
    
    if(nil != m_fetchedData)
        self.m_fetchedData = nil;
}

#pragma mark NSURLConnection Pinning methods

-(BOOL)connection:(NSURLConnection *)connection canAuthenticateAgainstProtectionSpace:(NSURLProtectionSpace *)space
{
    ASSERT(nil != connection);
    ASSERT(nil != space);
    
    return [[space authenticationMethod] isEqualToString: NSURLAuthenticationMethodServerTrust];
}

- (void)connection:(NSURLConnection *)connection didReceiveAuthenticationChallenge:(NSURLAuthenticationChallenge *)challenge
{
    ASSERT(nil != connection);
    ASSERT(nil != challenge);
    
    // Use the following to fetch the cert of interest. It will be in PEM format.
    // PEM format is (--BEGIN CERTIFICATE--, --END CERTIFICATE--).
    //     $ echo "Get HTTP/1.0" | openssl s_client -showcerts -connect www.random.org:443
    // Save the certifcate of interest to a file (for example, "random-org.pem").
    //   The certificate is the leaf, and should be located at certifcates[0] for
    //   non-ephemeral exchanges. Then, convert the certifcate to DER.
    //     $ openssl x509 -in "random-org.pem" -inform PEM –out "random-org.der" -outform DER
    
    if ([[[challenge protectionSpace] authenticationMethod] isEqualToString: NSURLAuthenticationMethodServerTrust])
    {
        do
        {
            SecTrustRef serverTrust = [[challenge protectionSpace] serverTrust];
            ASSERT(nil != serverTrust);
            if(!(nil != serverTrust)) break; /* failed */
            
            OSStatus status = SecTrustEvaluate(serverTrust, NULL);
            ASSERT(errSecSuccess == status);
            if(!(errSecSuccess == status)) break; /* failed */
            
            SecCertificateRef serverCertificate = SecTrustGetCertificateAtIndex(serverTrust, 0);
            ASSERT(nil != serverTrust);
            if(!(nil != serverTrust)) break;  /* failed */
            
            CFDataRef serverCertificateData = SecCertificateCopyData(serverCertificate);
            ASSERT(nil != serverCertificateData);
            if(!(nil != serverCertificateData)) break;  /* failed */
            
            [(id)serverCertificateData autorelease];
            const UInt8* const data = CFDataGetBytePtr(serverCertificateData);
            const CFIndex size = CFDataGetLength(serverCertificateData);
            
            ASSERT(nil != data);
            ASSERT(size > 0);
            if(!(nil != data) || !(size > 0)) break; /* failed */
            
            // (lldb) p data
            // (const UInt8 *) $0 = 0x1d8d3820
            // (lldb) p size
            // (CFIndex) $1 = 1772
            // (lldb) po serverCertificateData
            // $2 = 0x1d8d3800 <308206e8 308205d0 a0030201 02021074 b805ae19
            // e5ad4bed 4c3a20e6af02c930 0d06092a 864886f7 0d010105 05003081
            // 8e310b30 09060355 04061302 4742311b ... >
            
            NSData* cert1 = [NSData dataWithBytes:data length:(NSUInteger)size];
            ASSERT(nil != cert1);
            if(!(nil != cert1)) break; /* failed */
            
            NSString *file = [[NSBundle mainBundle] pathForResource:@"random-org" ofType:@"der"];
            ASSERT(nil != file);
            if(!(nil != file)) break; /* failed */
            
            NSData* cert2 = [NSData dataWithContentsOfFile:file];
            ASSERT(nil != cert2);
            if(!(nil != cert2)) break; /* failed */
            
            const BOOL equal = [cert1 isEqualToData:cert2];
            ASSERT(NO != equal);
            if(!(NO != equal)) break; /* failed */
            
            // The only good exit point
            return [[challenge sender] useCredential: [NSURLCredential credentialForTrust: serverTrust]
                          forAuthenticationChallenge: challenge];
            
        } while (0);
    }
    
    // Bad dog
    return [[challenge sender] cancelAuthenticationChallenge: challenge];
}

@end
